import warnings

from runeq.config import BaseConfig
from runeq.errors import InitializationError
from runeq.resources.client import (
    GraphClient,
    StreamClient,
    StriveClient,
    global_graph_client,
    global_stream_client,
    global_strive_client,
)
from runeq.resources.session.namespaces import (
    EventNamespace,
    OrgNamespace,
    PatientNamespace,
    ProjectNamespace,
    SleepNamespace,
    StreamMetadataNamespace,
    StreamNamespace,
    UserNamespace,
)


class Session:
    """
    Session object to access Rune APIs.
    """

    org: OrgNamespace
    event: EventNamespace
    patient: PatientNamespace
    project: ProjectNamespace
    sleep: SleepNamespace
    stream_metadata: StreamMetadataNamespace
    stream: StreamNamespace
    user: UserNamespace

    graph_client: GraphClient
    stream_client: StreamClient
    strive_client: StriveClient

    def __init__(
        self,
        config: BaseConfig,
    ):
        self._check_initialized()

        self.graph_client = GraphClient(config)
        self.stream_client = StreamClient(config)
        self.strive_client = StriveClient(config)

        self.org = OrgNamespace(self.graph_client)
        self.event = EventNamespace(self.graph_client)
        self.patient = PatientNamespace(self.graph_client)
        self.project = ProjectNamespace(self.graph_client)
        self.sleep = SleepNamespace(self.strive_client)
        self.stream_metadata = StreamMetadataNamespace(
            self.graph_client, self.stream_client
        )
        self.stream = StreamNamespace(self.stream_client)
        self.user = UserNamespace(self.graph_client)

    def _check_initialized(self):
        """
        Check if runeq has been globally initialized.
        """
        try:
            global_graph_client()
            global_stream_client()
            global_strive_client()
        except InitializationError:
            pass
        else:
            warnings.warn(
                "runeq has been globally initialized, and you are using a session-based client. You probably want to use one or the other.",
                stacklevel=2,
            )



import inspect
import os
from types import ModuleType
from typing import Callable
from unittest import TestCase

from runeq import Config, resources
from runeq.resources.session.session import Session

TEST_ROOT = os.path.dirname(os.path.realpath(__file__))

SKIP_MODULES = ["client", "common"]


class TestConfig(TestCase):
    """
    Test the Session object.
    """

    @classmethod
    def setUpClass(cls) -> None:
        cls.access_token_good_config = f"{TEST_ROOT}/data/access_token_good_config.yaml"

    def test_coverage(self):
        """
        Test the coverage of the Session object.
        """
        session = Session(Config(self.access_token_good_config))

        for module_name, module in inspect.getmembers(resources):
            if self.valid_module(module_name, module):
                for method_name, method in inspect.getmembers(module):
                    if self.valid_method(
                        method_name, method
                    ) and self.method_defined_in_module(method, module):
                        self.assertTrue(
                            hasattr(session, module_name),
                            f"Resource {module_name} not found in session",
                        )

                        namespace = getattr(session, module_name)
                        self.assertTrue(
                            hasattr(namespace, method_name),
                            f"Method {method_name} not found in session",
                        )

    def valid_module(self, module_name: str, module: ModuleType) -> bool:
        return (
            inspect.ismodule(module)
            and not module_name.startswith("_")
            and module_name not in SKIP_MODULES
        )

    def valid_method(self, method_name: str, method: Callable) -> bool:
        return inspect.isfunction(method) and not method_name.startswith("_")

    def method_defined_in_module(self, method: Callable, module: ModuleType) -> bool:
        return inspect.getmodule(method) == module



"""
Internal types and utilities. Not intended for external use.

"""

import time
from datetime import date, datetime
from typing import Union

_time_type = Union[int, float, date, datetime]


def _time_type_to_unix_secs(t: _time_type) -> Union[int, float]:
    """Standardize time input as timestamp (in unix secs)"""
    if isinstance(t, datetime):
        return t.timestamp()
    elif isinstance(t, date):
        return time.mktime(t.timetuple())
    else:
        return t



"""
Query sleep data directly from the Strive API.

"""

from datetime import date, timedelta
from typing import Dict, List, Optional

from .client import StriveClient, global_strive_client


def get_sleep_metrics(
    patient_id: str,
    start_date: date,
    end_date: date,
    client: Optional[StriveClient] = None,
) -> List[Dict]:
    """Fetch sleep metrics for a patient.

    Data is fetched from the Strive API in segments of 30 days at a time.

    Args:
        patient_id: The patient ID.
        start_date: The start date for the query.
        end_date: The end date for the query.
        client: Optional StriveClient instance. If not provided, uses the global client.
    """

    client = client or global_strive_client()
    all_metrics = []
    # fetch sleep in chunks of 30 days
    chunk_size = timedelta(days=30)
    chunk_start = start_date
    while chunk_start <= end_date:
        # the end is either the end of the next 30 days or the end date if
        # earlier
        chunk_end = min(chunk_start + chunk_size, end_date)
        resp = client.get(
            "/api/v3/sleep_metrics",
            params={
                "patient_id": patient_id,
                "start_date": chunk_start.strftime("%Y-%m-%d"),
                "end_date": chunk_end.strftime("%Y-%m-%d"),
            },
        )
        # if non-200, raise exception
        resp.raise_for_status()

        all_metrics.extend(resp.json()["data"]["sleep_metrics_healthkit"])
        chunk_start = chunk_end + timedelta(days=1)

    return all_metrics



"""
Tests for fetching strive data.

"""

from datetime import date
from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import StriveClient
from runeq.resources.sleep import get_sleep_metrics


class TestStriveData(TestCase):
    """
    Unit tests for the data queries in the strive module.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.strive_client = StriveClient(
            Config(client_key_id="test", client_access_key="config")
        )
        self.maxDiff = None

    @mock.patch("runeq.resources.client.requests.get")
    def test_get_sleep_metrics(self, mock_requests):
        """
        Test get sleep metrics for a given patient.

        """
        expected_data_dec = {
            "data": {
                "sleep_metrics_healthkit": [
                    {
                        "date": "2024-12-01",
                        "id": 1,
                        "qa_large_gaps": False,
                        "qa_stage_overlap": False,
                        "sleep_start_time": 1733108400.0,  # 2024-12-01 22:00:00
                        "sleep_end_time": 1733140800.0,  # 2024-12-02 07:00:00
                        "timezone": "America/New_York",
                        "total_sleep_time": 500,
                        "total_stage_time_core": 125,
                        "total_stage_time_deep": 250,
                        "total_stage_time_rem": 125,
                        "total_stage_time_unspecified": 125,
                        "waso": 25,
                        "number_of_awakenings": 2,
                    }
                ]
            }
        }
        expected_data_jan = {
            "data": {
                "sleep_metrics_healthkit": [
                    {
                        "date": "2025-01-09",
                        "id": 1,
                        "qa_large_gaps": False,
                        "qa_stage_overlap": False,
                        "sleep_start_time": 1736391600.0,  # 2025-01-08 22:00:00
                        "sleep_end_time": 1736427600.0,  # 2025-01-09 08:00:00
                        "timezone": "America/New_York",
                        "total_sleep_time": 500,
                        "total_stage_time_core": 125,
                        "total_stage_time_deep": 250,
                        "total_stage_time_rem": 125,
                        "total_stage_time_unspecified": 125,
                        "waso": 25,
                        "number_of_awakenings": 2,
                    }
                ]
            }
        }

        expected_metrics = [
            {
                "date": "2024-12-01",
                "id": 1,
                "qa_large_gaps": False,
                "qa_stage_overlap": False,
                "sleep_start_time": 1733108400.0,
                "sleep_end_time": 1733140800.0,
                "timezone": "America/New_York",
                "total_sleep_time": 500,
                "total_stage_time_core": 125,
                "total_stage_time_deep": 250,
                "total_stage_time_rem": 125,
                "total_stage_time_unspecified": 125,
                "waso": 25,
                "number_of_awakenings": 2,
            },
            {
                "date": "2025-01-09",
                "id": 1,
                "qa_large_gaps": False,
                "qa_stage_overlap": False,
                "sleep_start_time": 1736391600.0,
                "sleep_end_time": 1736427600.0,
                "timezone": "America/New_York",
                "total_sleep_time": 500,
                "total_stage_time_core": 125,
                "total_stage_time_deep": 250,
                "total_stage_time_rem": 125,
                "total_stage_time_unspecified": 125,
                "waso": 25,
                "number_of_awakenings": 2,
            },
        ]

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.json.return_value = expected_data_dec

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.json.return_value = expected_data_jan

        mock_requests.side_effect = [mock_response1, mock_response2]
        actual = get_sleep_metrics(
            "test_patient_id",
            start_date=date(2024, 12, 1),
            end_date=date(2025, 1, 10),
            client=self.strive_client,
        )
        expected = expected_metrics
        self.assertEqual(expected, actual)
        self.assertEqual(mock_requests.call_count, 2)



"""
Fetch data from Rune APIs.

By default, globally-initialized clients are used for all API requests (see
:class:`~runeq.resources.client.initialize`). Functions that make API requests
also accept optional client(s), which can be used in lieu of the global
initialization.

Metadata is fetched from Rune's GraphQL API
(https://graph.runelabs.io/graphql), using a
:class:`~runeq.resources.client.GraphClient`. Timeseries data is fetched
from the `V2 Stream API <https://docs.runelabs.io/stream/v2/index.html>`_,
using a :class:`~runeq.resources.client.StreamClient`.

"""

from . import (
    client,
    common,
    event,
    org,
    patient,
    project,
    sleep,
    stream,
    stream_metadata,
    user,
)
from .session import Session

__all__ = [
    "client",
    "common",
    "event",
    "org",
    "patient",
    "stream",
    "stream_metadata",
    "user",
    "sleep",
    "project",
    "Session",
]



import datetime
from datetime import date
from typing import Dict, Iterable, Iterator, List, Literal, Optional, Union

import pandas as pd

from runeq.resources.client import GraphClient, StreamClient, StriveClient
from runeq.resources.event import (
    EventSet,
    get_patient_activity_events,
    get_patient_events,
    get_patient_medication_events,
    get_patient_symptom_events,
    get_patient_wellbeing_events,
)
from runeq.resources.internal import _time_type
from runeq.resources.org import Org, OrgSet, get_org, get_orgs, set_active_org
from runeq.resources.patient import (
    Device,
    DeviceSet,
    Patient,
    PatientSet,
    get_all_devices,
    get_all_patients,
    get_device,
    get_patient,
    get_patient_devices,
)
from runeq.resources.project import (
    Project,
    ProjectPatientMetadataSet,
    ProjectSet,
    get_cohort_patients,
    get_project,
    get_project_patients,
    get_projects,
)
from runeq.resources.sleep import get_sleep_metrics
from runeq.resources.stream import (
    get_stream_aggregate_window,
    get_stream_availability,
    get_stream_daily_aggregate,
    get_stream_data,
)
from runeq.resources.stream_metadata import (
    StreamMetadata,
    StreamMetadataSet,
    StreamTypeSet,
    get_all_stream_types,
    get_patient_stream_metadata,
    get_stream_availability_dataframe,
    get_stream_dataframe,
    get_stream_metadata,
)
from runeq.resources.user import User, get_current_user


class OrgNamespace:
    client: GraphClient

    def __init__(self, client: GraphClient):
        self.client = client

    def get_org(self, org_id: str) -> Org:
        return get_org(org_id, client=self.client)

    def get_orgs(self) -> OrgSet:
        return get_orgs(client=self.client)

    def set_active_org(self, org: Union[str, Org]) -> Org:
        return set_active_org(org, client=self.client)


class EventNamespace:
    client: GraphClient

    def __init__(self, client: GraphClient):
        self.client = client

    def get_patient_events(
        self,
        patient_id: str,
        start_time: _time_type,
        end_time: _time_type,
        include_filters: Optional[list] = None,
    ) -> EventSet:
        return get_patient_events(
            patient_id,
            start_time,
            end_time,
            include_filters=include_filters,
            client=self.client,
        )

    def get_patient_activity_events(
        self,
        patient_id: str,
        start_time: _time_type,
        end_time: _time_type,
    ) -> EventSet:
        return get_patient_activity_events(
            patient_id,
            start_time,
            end_time,
            client=self.client,
        )

    def get_patient_medication_events(
        self,
        patient_id: str,
        start_time: _time_type,
        end_time: _time_type,
    ) -> EventSet:
        return get_patient_medication_events(
            patient_id,
            start_time,
            end_time,
            client=self.client,
        )

    def get_patient_symptom_events(
        self,
        patient_id: str,
        start_time: _time_type,
        end_time: _time_type,
    ) -> EventSet:
        return get_patient_symptom_events(
            patient_id,
            start_time,
            end_time,
            client=self.client,
        )

    def get_patient_wellbeing_events(
        self,
        patient_id: str,
        start_time: _time_type,
        end_time: _time_type,
    ) -> EventSet:
        return get_patient_wellbeing_events(
            patient_id,
            start_time,
            end_time,
            client=self.client,
        )


class PatientNamespace:
    client: GraphClient

    def __init__(self, client: GraphClient):
        self.client = client

    def get_patient(self, patient_id: str) -> Patient:
        return get_patient(patient_id, client=self.client)

    def get_all_patients(self) -> PatientSet:
        return get_all_patients(client=self.client)

    def get_device(
        self,
        patient: Union[Patient, str],
        device_id: str,
    ) -> Device:
        return get_device(patient, device_id, client=self.client)

    def get_patient_devices(
        self,
        patient: Union[Patient, str],
    ) -> DeviceSet:
        return get_patient_devices(patient, client=self.client)

    def get_all_devices(
        self,
        patients: Union[PatientSet, List[str]] = None,
    ) -> DeviceSet:
        return get_all_devices(patients, client=self.client)


class ProjectNamespace:
    client: GraphClient

    def __init__(self, client: GraphClient):
        self.client = client

    def get_project(self, project_id: str) -> Project:
        return get_project(project_id, client=self.client)

    def get_projects(self) -> ProjectSet:
        return get_projects(client=self.client)

    def get_project_patients(
        self,
        project_id: str,
    ) -> ProjectPatientMetadataSet:
        return get_project_patients(project_id, client=self.client)

    def get_cohort_patients(
        self,
        cohort_id: str,
    ) -> ProjectPatientMetadataSet:
        return get_cohort_patients(cohort_id, client=self.client)


class SleepNamespace:
    client: StriveClient

    def __init__(self, client: StriveClient):
        self.client = client

    def get_sleep_metrics(
        self,
        patient_id: str,
        start_date: date,
        end_date: date,
    ) -> List[Dict]:
        return get_sleep_metrics(
            patient_id,
            start_date,
            end_date,
            client=self.client,
        )


class StreamMetadataNamespace:
    graph_client: GraphClient
    stream_client: StreamClient

    def __init__(self, graph_client: GraphClient, stream_client: StreamClient):
        self.graph_client = graph_client
        self.stream_client = stream_client

    def get_all_stream_types(self) -> StreamTypeSet:
        return get_all_stream_types(client=self.graph_client)

    def get_stream_metadata(
        self,
        stream_ids: Union[str, Iterable[str]],
    ) -> Union[StreamMetadata, StreamMetadataSet]:
        return get_stream_metadata(stream_ids, client=self.graph_client)

    def get_patient_stream_metadata(
        self,
        patient_id: str,
        device_id: Optional[str] = None,
        stream_type_id: Optional[str] = None,
        algorithm: Optional[str] = None,
        category: Optional[str] = None,
        measurement: Optional[str] = None,
        **parameters,
    ) -> StreamMetadataSet:
        return get_patient_stream_metadata(
            patient_id,
            device_id=device_id,
            stream_type_id=stream_type_id,
            algorithm=algorithm,
            category=category,
            measurement=measurement,
            client=self.graph_client,
            **parameters,
        )

    def get_stream_dataframe(
        self,
        stream_ids: Union[str, Iterable[str]],
        start_time: Optional[Union[float, datetime.date]] = None,
        start_time_ns: Optional[int] = None,
        end_time: Optional[Union[float, datetime.date]] = None,
        end_time_ns: Optional[int] = None,
        limit: Optional[int] = None,
        page_token: Optional[str] = None,
        timestamp: Optional[str] = "iso",
        timezone: Optional[int] = None,
        translate_enums: Optional[bool] = True,
    ) -> pd.DataFrame:
        return get_stream_dataframe(
            stream_ids,
            start_time=start_time,
            start_time_ns=start_time_ns,
            end_time=end_time,
            end_time_ns=end_time_ns,
            limit=limit,
            page_token=page_token,
            timestamp=timestamp,
            timezone=timezone,
            translate_enums=translate_enums,
            stream_client=self.stream_client,
            graph_client=self.graph_client,
        )

    def get_stream_availability_dataframe(
        self,
        stream_ids: Union[str, Iterable[str]],
        start_time: Union[float, datetime.date],
        end_time: Union[float, datetime.date],
        resolution: int,
        batch_operation: Optional[str] = None,
        limit: Optional[int] = None,
        page_token: Optional[str] = None,
        timestamp: Optional[str] = "iso",
        timezone: Optional[int] = None,
    ) -> pd.DataFrame:
        return get_stream_availability_dataframe(
            stream_ids,
            start_time=start_time,
            end_time=end_time,
            resolution=resolution,
            batch_operation=batch_operation,
            limit=limit,
            page_token=page_token,
            timestamp=timestamp,
            timezone=timezone,
            stream_client=self.stream_client,
            graph_client=self.graph_client,
        )


class StreamNamespace:
    client: StreamClient

    def __init__(self, client: StreamClient):
        self.client = client

    def get_stream_data(
        self,
        stream_id: str,
        start_time: Optional[_time_type] = None,
        start_time_ns: Optional[int] = None,
        end_time: Optional[_time_type] = None,
        end_time_ns: Optional[int] = None,
        format: Optional[str] = "csv",
        limit: Optional[int] = None,
        page_token: Optional[str] = None,
        timestamp: Optional[str] = "iso",
        timezone: Optional[int] = None,
        timezone_name: Optional[str] = None,
        translate_enums: Optional[bool] = True,
    ) -> Iterator[Union[str, dict]]:
        return get_stream_data(
            stream_id,
            start_time=start_time,
            start_time_ns=start_time_ns,
            end_time=end_time,
            end_time_ns=end_time_ns,
            format=format,
            limit=limit,
            page_token=page_token,
            timestamp=timestamp,
            timezone=timezone,
            timezone_name=timezone_name,
            translate_enums=translate_enums,
            client=self.client,
        )

    def get_stream_availability(
        self,
        stream_ids: Union[str, Iterable[str]],
        start_time: _time_type,
        end_time: _time_type,
        resolution: int,
        batch_operation: Optional[str] = None,
        format: Optional[str] = "csv",
        limit: Optional[int] = None,
        page_token: Optional[str] = None,
        timestamp: Optional[str] = "iso",
        timezone: Optional[int] = None,
        timezone_name: Optional[str] = None,
    ) -> Iterator[Union[str, dict]]:
        return get_stream_availability(
            stream_ids,
            start_time=start_time,
            end_time=end_time,
            resolution=resolution,
            batch_operation=batch_operation,
            format=format,
            limit=limit,
            page_token=page_token,
            timestamp=timestamp,
            timezone=timezone,
            timezone_name=timezone_name,
            client=self.client,
        )

    def get_stream_daily_aggregate(
        self,
        stream_id: str,
        start_time: _time_type,
        resolution: int,
        n_days: int,
    ) -> dict:
        return get_stream_daily_aggregate(
            stream_id,
            start_time,
            resolution,
            n_days,
            client=self.client,
        )

    def get_stream_aggregate_window(
        self,
        stream_id: str,
        start_time: _time_type,
        end_time: _time_type,
        resolution: int,
        aggregate_function: Literal["sum", "mean"],
        timestamp: Optional[str] = "iso",
        timezone: Optional[int] = None,
        timezone_name: Optional[str] = None,
    ) -> dict:
        return get_stream_aggregate_window(
            stream_id,
            start_time=start_time,
            end_time=end_time,
            resolution=resolution,
            aggregate_function=aggregate_function,
            timestamp=timestamp,
            timezone=timezone,
            timezone_name=timezone_name,
            client=self.client,
        )


class UserNamespace:
    client: GraphClient

    def __init__(self, client: GraphClient):
        self.client = client

    def get_current_user(self) -> User:
        return get_current_user(client=self.client)
